import os, sys, pygame, operator
from pygame.locals import *
from random import Random
from time import time

#--------0---------0---------0---------0---------0---------0---------0---------0
def load_png( name, extradir=''):
	fullname = os.path.join('data', extradir, name)
	try:
		image = pygame.image.load(fullname)
		if image.get_alpha is None:
			image = image.convert()
		else:
			image = image.convert_alpha()
	except pygame.error, message:
		print 'Cannot load png:', fullname
		raise SystemExit, message
	return image, image.get_rect()


#--------0---------0---------0---------0---------0---------0---------0---------0
def GetMoveStateFor(fromRect,toRect):
	print "from", fromRect
	print "to", toRect
	rise = toRect.centery - fromRect.centery
	run  = toRect.centerx - fromRect.centerx

	print "rise ", rise, " run ", run

	absRise = abs(rise)
	absRun = abs(run)
	if absRise > absRun:
		rise = rise/absRise
		run = float(run)/absRise
	else:
		run = run/absRun
		rise = float(rise)/absRun

	print "calculated move: ", [run,rise]
	return [run,rise]

#--------0---------0---------0---------0---------0---------0---------0---------0
class Player:
	def __init__(self, game):
		self.game = game
		self.money = 30
		self.antibioticCounter = 0

	def Tick(self):
		#every 10 ticks, you have to buy antibiotics & growth hormone
		self.antibioticCounter = self.antibioticCounter +1
		if self.antibioticCounter > 9:
			self.antibioticCounter = 0
			charge = 10 + 5 * len(self.game.sheepGroup)
			self.SetMoney( self.money - charge )

	def SetMoney(self, amt):
		self.money = amt
		if self.money < 0:
			self.game.End( "You're Fired!  You ran out of money!" )

#--------0---------0---------0---------0---------0---------0---------0---------0
#--------0---------0---------0---------0---------0---------0---------0---------0
class SheepGame:
	def __init__(self, screen, allSprites ):
		self.screen = screen
		self.time = int(time())
		self.selectedSprite = None

		self.allowSignalsThrough = 1
		self.paused = 0
		self.sheepQueue = []

		self.allSprites = allSprites

		self.hudGroup = pygame.sprite.Group()
		self.grassGroup = RectSmartGroup()
		self.sheepGroup = pygame.sprite.Group()
		self.fenceGroup = pygame.sprite.Group()
		self.mapGroup = pygame.sprite.Group()
		self.labGroup = pygame.sprite.RenderUpdates()
		self.gameOverGroup = pygame.sprite.RenderUpdates()

		self.player1 = Player(self)

		self.map = Map(self, screen, self.player1)
		self.lab = Lab(self, screen, self.player1)

		self.boundingRect = self.screen.get_rect()
		self.boundingRect.top = self.boundingRect.top + 80
		self.boundingRect.height = self.boundingRect.height - 80

		self.grassGrowthCounter = 0
		self.grassGrowthRate = 3

		gb = GrassBloom(self.boundingRect, self)
		fence = Fence(self, [770,100])
		fence = Fence(self, [770,150])
		fence = Fence(self, [770,200])
		fence = Fence(self, [770,250])
		fence = Fence(self, [770,300])
		fence = Fence(self, [770,400])
		fence = Fence(self, [770,550])
		fence = Fence(self, [770,600])
		self.MakeSheepOfClass( Sheep, 1 )
		moneySprite = MoneySprite(self, screen, self.player1)
		#sheep = Sheep(self, self.boundingRect, self.player1 )

		self.mode = 'field1'

	#----------------------------------------------------------------------
	def End( self, message ):
		print "GAME ENDS NOW!!!"
		#font = pygame.font.Font(None, 60 )
		#over = pygame.sprite.Sprite()
		over = HUDItem(self, self.screen, message)
		#over.image = font.render(message, 1, (0,0,0) )
		#over.rect = over.image.get_rect()
		over.rect.center= (400,300)

		self.mode = 'over'
		self.paused = 1

	#----------------------------------------------------------------------
	def MakeSheepOfClass(self, sheepClass, delay=10):
		sheep = sheepClass(self,self.boundingRect,self.player1)
		self.EnqueueSheep( sheep, delay )

	#----------------------------------------------------------------------
	def GetActiveGroups(self):
		if self.mode == 'field1':
			return self.allSprites
		elif self.mode == 'lab':
			return self.labGroup
		elif self.mode == 'over':
			return self.gameOverGroup

	#----------------------------------------------------------------------
	def Tick(self):
		#every second
		if self.paused:
			return
		if int(time()) - self.time > 1:
			self.time = int(time())
			self.player1.Tick()
			self.map.update()
			if len(self.grassGroup) < 50:
				self.grassGrowthCounter = self.grassGrowthCounter + 1
				if self.grassGrowthCounter >= self.grassGrowthRate:
					self.grassGrowthCounter = 0
					gb = GrassBloom(self.boundingRect, self)
			#gotta iterate backwards so deletions dont' affect order
			#print self.sheepQueue
			for i in range(len(self.sheepQueue)-1,-1,-1):
				if self.sheepQueue[i][0] <= self.time:
					self.sheepQueue[i][1].Birth()
					del self.sheepQueue[i]


	#----------------------------------------------------------------------
	def RegisterSprite(self, newSprite):
		if isinstance(newSprite, HUDItem):
			self.hudGroup.add( newSprite )
			self.allSprites.add( newSprite )
			self.labGroup.add( newSprite )
			self.gameOverGroup.add( newSprite )
		if isinstance(newSprite, Grass):
			self.grassGroup.add( newSprite )
			self.allSprites.add( newSprite )
		if isinstance(newSprite, Fence):
			self.fenceGroup.add( newSprite )
			self.allSprites.add( newSprite )
		elif isinstance(newSprite, Sheep):
			self.sheepGroup.add( newSprite )
			self.allSprites.add( newSprite )
		elif newSprite.__class__ == Door \
		   or newSprite.__class__ == OpenLabButton:
			self.mapGroup.add( newSprite )
			self.allSprites.add( newSprite )
		elif newSprite.__class__ == LabButton:
			self.labGroup.add( newSprite )
			print "lab!"
		else:
			#unknown sprite?!?!
			self.allSprites.add( newSprite )


	#----------------------------------------------------------------------
	def SelectSprite(self, sprite):
		self.selectedSprite = sprite

	#----------------------------------------------------------------------
	def UnSelectSprite(self, sprite):
		if self.selectedSprite == sprite:
			self.selectedSprite = None

	#----------------------------------------------------------------------
	def EnqueueSheep(self, sheep, delay=10):
		self.sheepQueue.append( [self.time+delay, sheep] )
		
	#----------------------------------------------------------------------
	def OpenLab(self):
		print "SHOULD OPEN LAB HERE"
		self.mode = 'lab'
		self.paused = 1
		bgSurface, devnull = load_png('lab.png')
		self.screen.blit(bgSurface, (0,0) )
		pygame.display.flip()

	#----------------------------------------------------------------------
	def CloseLab(self):
		print "SHOULD CLOSE LAB HERE"
		self.mode = 'field1'
		self.paused = 0
		bgSurface, devnull = load_png('backgr.png')
		self.screen.blit(bgSurface, (0,0) )
		pygame.display.flip()



	#----------------------------------------------------------------------
	def UserEvent(self, event):
		if self.allowSignalsThrough and self.mode == 'field1':
			for sheep in self.sheepGroup.sprites():
				sheep.UserEvent( event, self.selectedSprite )
			for sprite in self.mapGroup.sprites():
				sprite.UserEvent( event, self.selectedSprite )
		elif self.allowSignalsThrough and self.mode == 'lab':
			for sprite in self.labGroup.sprites():
				sprite.UserEvent( event, self.selectedSprite )

#--------0---------0---------0---------0---------0---------0---------0---------0
class Lab:
	def __init__(self, game, screen, player1):
		self.screen = screen
		self.game = game
		self.player1 = player1
		self.exitButton= LabButton(self.game, self, "Exit", [100,20],
		                           "",
		                           self.ClickExit, 60)
		self.sRButton= LabButton(self.game, self, "Regular GM", 
		                           [530,410],
		                           "",
		                           self.ClickSheepR)
		self.sHYButton= LabButton(self.game, self, "High Yield Genes", 
		                           [530,490],
		                           "",
		                           self.ClickSheepHY)
	
	
	
	def ClickExit(self):
		self.game.CloseLab()

	def ClickSheepR(self):
		cost = 100
		if self.player1.money < cost:
			return
		self.player1.money = self.player1.money - cost
		self.game.MakeSheepOfClass( Sheep )

	def ClickSheepHY(self):
		cost = 200
		if self.player1.money < cost:
			return
		self.player1.money = self.player1.money - cost
		self.game.MakeSheepOfClass( HighYieldSheep )
		

#--------0---------0---------0---------0---------0---------0---------0---------0
class LabButton(pygame.sprite.Sprite):
	def __init__(self, game, lab, text, position, desc='', 
	             clickFn=None, fontSize=30):
		pygame.sprite.Sprite.__init__(self)
		self.game = game
		self.lab = lab
		self.desc = desc
		self.clickFn = clickFn

		font = pygame.font.Font(None, fontSize )
		self.image = font.render(text, 1, (0,0,0) )
		self.rect = self.image.get_rect()
		self.rect.topleft = position

		self.game.RegisterSprite(self)

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		if self.clickFn is not None and event.type == MOUSEBUTTONDOWN:
			clickBox = self.rect.move(0,0)
			if clickBox.collidepoint(pygame.mouse.get_pos()):
				self.clickFn()



#--------0---------0---------0---------0---------0---------0---------0---------0
class HUDItem(pygame.sprite.Sprite):
	def __init__(self, game, screen, text):
		pygame.sprite.Sprite.__init__(self)
		self.screen = screen
		self.game = game

		font = pygame.font.Font(None, 60 )
		self.image = font.render(text, 1, (0,0,0) )
		self.rect = self.image.get_rect()
		self.rect.topleft = (650,10)

		self.game.RegisterSprite(self)

	#----------------------------------------------------------------------
	def update(self):
		pass
		#moneyStr = "$" + str(self.player.money)
		#font = pygame.font.Font(None, 60 )
		#self.image = font.render(moneyStr, 1, (0,0,0) )
		#self.rect = self.image.get_rect()
		#self.rect.topleft = (650,10)

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		pass


#--------0---------0---------0---------0---------0---------0---------0---------0
class MoneySprite(HUDItem):
	def __init__(self, game, screen, player):
		self.player = player
		moneyStr = "$" + str(self.player.money)
		HUDItem.__init__(self,game,screen,moneyStr)

	#----------------------------------------------------------------------
	def update(self):
		moneyStr = "$" + str(self.player.money)
		font = pygame.font.Font(None, 60 )
		self.image = font.render(moneyStr, 1, (0,0,0) )

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		pass
		

#--------0---------0---------0---------0---------0---------0---------0---------0
class Map:
	def __init__(self, game, screen, player1):
		self.screen = screen
		self.game = game
		self.player1 = player1
		self.door = Door(game)
		self.labButton = OpenLabButton(game)

		#self.moneySprite = pygame.sprite.Sprite()
		#moneyStr = "$" + str(self.player1.money)
		#font = pygame.font.Font(None, 60 )
		#self.moneySprite.image = font.render(moneyStr, 1, (0,0,0) )
		#self.moneySprite.rect = self.moneySprite.image.get_rect()
		#self.moneySprite.rect.topleft = (650,10)

		#self.game.RegisterSprite(self.moneySprite)


	def update(self):
		#moneyStr = "$" + str(self.player1.money)
		#font = pygame.font.Font(None, 60 )
		#self.moneySprite.image = font.render(moneyStr, 1, (0,0,0) )
		#self.moneySprite.rect = self.moneySprite.image.get_rect()
		#self.moneySprite.rect.topleft = (650,10)
		pass
		

#--------0---------0---------0---------0---------0---------0---------0---------0
class OpenLabButton(pygame.sprite.Sprite):
	def __init__(self, game):
		pygame.sprite.Sprite.__init__(self)
		self.game = game
		self.image, self.rect = load_png( 'door.png' )
		self.rect.center = [60,40]

		game.RegisterSprite(self)

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		if event.type == MOUSEBUTTONDOWN:
			clickBox = self.rect.move(0,0)
			if clickBox.collidepoint(pygame.mouse.get_pos()):
				if selectedSprite is not None:
					selectedSprite.ActOn( self )

				if self.game.paused:
					self.game.CloseLab()
				else:
					self.game.OpenLab()


#--------0---------0---------0---------0---------0---------0---------0---------0
class Door(pygame.sprite.Sprite):
	def __init__(self, game):
		pygame.sprite.Sprite.__init__(self)
		self.game = game
		self.image, self.rect = load_png( 'door.png' )
		self.rect.center = [300,90]

		game.RegisterSprite(self)

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		if event.type == MOUSEBUTTONDOWN:
			clickBox = self.rect.inflate(-10,-10)
			if clickBox.collidepoint(pygame.mouse.get_pos()):
				if selectedSprite is not None:
					selectedSprite.ActOn( self )



#--------0---------0---------0---------0---------0---------0---------0---------0
class GrassBloom:
	def __init__(self, boundingRect, game):
		#decide where I'll sprout up
		bRect = boundingRect.inflate( -70, -70 )

		print bRect
		w = bRect.width
		h = bRect.height
		x = int(rng.normalvariate( w/2, w/4 )) % w
		y = int(rng.normalvariate( h/2, h/4 )) % h

		#print x, y
		
		#that gives us the x,y INSIDE the bBox, now we 
		#need the ABSOLUTE x and y
		x = bRect.left + x
		y = bRect.top + y

		#print x, y

		displacement = rng.randint(20,65)

		newGrass = Grass(game)
		newGrass.rect.center = (x+displacement,y)
		newGrass = Grass(game)
		newGrass.rect.center = (x-displacement,y+displacement-20)
		newGrass = Grass(game)
		newGrass.rect.center = (x,y-displacement)
		#print "Making a new grass sprite at ", x," ",y



#--------0---------0---------0---------0---------0---------0---------0---------0
class Grass(pygame.sprite.Sprite):
	def __init__(self, game):
		pygame.sprite.Sprite.__init__(self)
		self.image, self.rect = load_png( 'grass.png' )
		self.health = 100

		game.RegisterSprite(self)

	#----------------------------------------------------------------------
	def TakeBite(self):
		self.health = self.health - 50
		if self.health <= 0:
			self.KillAllSprites()

	#----------------------------------------------------------------------
	def KillAllSprites(self):
		self.kill()
		garbage.add( self )



#--------0---------0---------0---------0---------0---------0---------0---------0
class Fence(pygame.sprite.Sprite):
	def __init__(self, game, position):
		pygame.sprite.Sprite.__init__(self)
		self.game = game
		self.image, self.rect = load_png( 'fence.png' )
		self.health = 100

		self.rect.center = position
		self.rect.inflate_ip( -10, 0 )

		game.RegisterSprite(self)

#--------0---------0---------0---------0---------0---------0---------0---------0
class Sheep(pygame.sprite.Sprite):
	def __init__(self, game, boundingRect, owner, prefix=''):
		pygame.sprite.Sprite.__init__(self)
		self.game = game
		self.owner = owner
		self.health = 20
		self.selected = 0
		self.image, self.rect = load_png( prefix +'sheep.png' )
		self.rect.inflate_ip( -10, -10 )
		self.movestate = [0,0]
		self.speed = [10,10]
		self.boundingRect = boundingRect

		self.offImage = self.image
		self.onImage,d = load_png( prefix+'sheepselected.png')
		self.fullImage,d = load_png( prefix+'sheepfull.png')

		self.stallCounter = 0


		self.target = None

		self.mode = 'wander'

		self.growthRate = 20
		self.maxHealth  = 100

	#----------------------------------------------------------------------
	def Birth(self):
		self.rect.left = 400
		self.rect.top = 90
		self.mode = 'wander'
		self.movestate = [0,0]
		self.game.RegisterSprite(self)
		

	#----------------------------------------------------------------------
	def ChangeImage(self):
		if self.selected:
			self.image = self.onImage
		elif self.health >= self.maxHealth:
			self.image = self.fullImage
		else:
			self.image = self.offImage

	#----------------------------------------------------------------------
	def update(self):

		if self.mode == 'stalled':
			#TODO: this is kinda dirty maybe a timer would be better
			self.stallCounter = self.stallCounter + 1
			if self.stallCounter > 100:
				self.stallCounter = 0
				self.mode = 'wander'

		grasses = self.game.grassGroup
		hitbox = self.rect.inflate( -20, -20 )
		grassCollider = grasses.GetFirstSpriteThatCollides( hitbox )
		#colliders = pygame.sprite.spritecollide( self,grasses,0)
		if self.mode == 'wander':
			if grassCollider is not None:
				self.mode = 'stalled'
				#grass = colliders.pop()
				grassCollider.TakeBite()
				if self.health < self.maxHealth:
					self.health = self.health + self.growthRate
				else:
					self.ChangeImage()
			else:
				if rng.randint(1,6) == 1:
					self.movestate[0] = rng.randint(-1,1)
					self.movestate[1] = rng.randint(-1,1)
				self.Move()
		elif self.mode == 'shearing':
			hitbox = self.rect.move(0,0)
			if hitbox.colliderect( self.target.rect ):
				self.GetSheared()
			self.Move()

	#----------------------------------------------------------------------
	def KillAllSprites(self):
		self.kill()
		garbage.add( self )

	#----------------------------------------------------------------------
	def GetSheared(self):
		self.KillAllSprites()
		self.game.EnqueueSheep( self )
		self.owner.money = self.owner.money + self.health
		self.health = 0
		self.ChangeImage()
		

	#----------------------------------------------------------------------
	def Move(self):
		change = map(operator.mul, self.movestate, self.speed )
		if change == [0,0]:
			return

		self.oldPos = self.rect.move( (0,0) )
		self.rect.move_ip( change )
		
		limiter = self.Blocked( )
		if limiter.top is not None:
			self.rect.top    =limiter.top
			self.movestate[1] = 1
		if limiter.right is not None:
			self.rect.right  =limiter.right
		if limiter.bottom is not None:
			self.rect.bottom =limiter.bottom
			self.movestate[1] = -1
		if limiter.left is not None:
			self.rect.left   =limiter.left

	#----------------------------------------------------------------------
	def Blocked(self):
		limiter = BoundingBox()
		
		if self.rect.top    <= self.boundingRect.top:
			limiter.top    = self.boundingRect.top
		if self.rect.right  >= self.boundingRect.right:
			limiter.right  = self.boundingRect.right
		if self.rect.bottom >= self.boundingRect.bottom:
			limiter.bottom = self.boundingRect.bottom
		if self.rect.left   <= self.boundingRect.left:
			limiter.left   = self.boundingRect.left

		colliders = pygame.sprite.spritecollide( self, 
		                                       self.game.fenceGroup, 0 )

		if len(colliders) > 0:
			for sprite in colliders:

				if( limiter.top is None \
				    or sprite.rect.bottom > limiter.top ) \
				    and self.rect.top < sprite.rect.bottom \
				    and self.oldPos.top >= sprite.rect.bottom:
					limiter.top = sprite.rect.bottom
					limiter.topSprite = sprite

				if( limiter.right is None \
				    or sprite.rect.left < limiter.right ) \
				    and self.rect.right > sprite.rect.left \
				    and self.oldPos.right <= sprite.rect.left:
					limiter.right = sprite.rect.left
					limiter.rightSprite = sprite

				if( limiter.bottom is None \
				    or sprite.rect.top < limiter.bottom ) \
				    and self.rect.bottom > sprite.rect.top \
				    and self.oldPos.bottom <=sprite.rect.top:
					limiter.bottom = sprite.rect.top
					limiter.bottomSprite = sprite

				if( limiter.left is None \
				    or sprite.rect.right > limiter.left ) \
				    and self.rect.left < sprite.rect.right \
				    and self.oldPos.left >= sprite.rect.right:
					limiter.left = sprite.rect.right
					limiter.leftSprite = sprite

		return limiter

	#----------------------------------------------------------------------
	def UserEvent(self, event, selectedSprite):
		if event.type == MOUSEBUTTONDOWN:
			clickBox = self.rect.inflate(-10,-10)
			if clickBox.collidepoint(pygame.mouse.get_pos()):
				print "I got a user event!  I am a", \
				       self.__class__
				if selectedSprite is None:
					self.game.SelectSprite( self )
					self.selected = 1
					self.ChangeImage()
				else:
					selectedSprite.ActOn( self )

	#----------------------------------------------------------------------
	def ActOn(self, sprite):
		print "acting on a ", sprite
		if isinstance( sprite, Door ):
			self.mode = 'shearing'
			self.movestate = GetMoveStateFor(self.rect,sprite.rect)
			self.target = sprite
		else:
			print "I don't know what to do to a ", sprite.__class__
		self.game.UnSelectSprite( self )
		self.selected = 0
		self.ChangeImage()
		


#--------0---------0---------0---------0---------0---------0---------0---------0
class HighYieldSheep(Sheep):
	def __init__(self, game, boundingRect, owner):
		Sheep.__init__(self,game,boundingRect,owner, 'hy_')
		self.growthRate = 15
		self.maxHealth = 200

#--------0---------0---------0---------0---------0---------0---------0---------0
class BoundingBox:
	def __init__(self):
		self.top    = None
		self.right  = None
		self.bottom = None
		self.left   = None

		self.topSprite    = None
		self.rightSprite  = None
		self.bottomSprite = None
		self.leftSprite   = None

#--------0---------0---------0---------0---------0---------0---------0---------0
class RectSmartGroup( pygame.sprite.RenderClear ):
	def GetAllRects(self):
		rectlist = []
		for sprite in self.sprites():
			rectlist.append( sprite.rect )
		return rectlist

	def GetFirstSpriteThatCollides(self, target):
		for sprite in self.sprites():
			if sprite.rect.colliderect( target ):
				return sprite
		return None


#--------0---------0---------0---------0---------0---------0---------0---------0
def main():

	#Random Number Generator
	global rng
	rng = Random()

	pygame.init()
	screen = pygame.display.set_mode((800, 640))
	pygame.display.set_caption('Sheep Game')

	bgSurface, devnull = load_png('backgr.png')
	screen.blit(bgSurface, (0,0) )
	pygame.display.flip()

	clock = pygame.time.Clock()

	allSprites = pygame.sprite.RenderUpdates()
	global garbage
	garbage = RectSmartGroup()

	game = SheepGame( screen, allSprites )

	oldFPS = 0
	while 1:
		clock.tick(60)
		newFPS = int(clock.get_fps())
		if oldFPS is not newFPS:
			print "FPS: ", newFPS
			oldFPS = newFPS

		for event in pygame.event.get():
			if event.type == QUIT \
			  or (event.type == KEYDOWN and event.key == K_ESCAPE):
				return
			elif event.type == MOUSEBUTTONDOWN:
				game.UserEvent( event )

		activeGroup = game.GetActiveGroups()
		game.Tick()

		#updates
		activeGroup.update()

		#clears
		activeGroup.clear( screen, bgSurface )

		#Garbage has to be emptied each time
		garbage.clear( screen, bgSurface )
		garbageRects = garbage.GetAllRects()
		garbage.empty()


		#draws (note, order is important)
		activeGroupRects = activeGroup.draw(screen)

		pygame.display.update( activeGroupRects \
		                     + garbageRects
				     )


if __name__ == '__main__': main()
